import numpy as np
import cv2 as cv
import matplotlib.pyplot as plt
from math import pi as PI
from black_and_white import BlackWhiteTransform
import logging

logger = logging.getLogger(__name__)

# For debugging purposes
gsize = None
ginertia = None
ls = []

class Colors:
    RED = "RED"
    BLUE = "BLUE"
    GREY = "GREY" 

def debug(image):
    cv.imshow("debug", image)
    cv.waitKey(0)

def debug_contours(contours, image):
    mask = np.zeros((image.shape[0], image.shape[1]))
    cv.drawContours(mask, contours, -1, (255), 1)
    cv.imshow("debug", mask)
    cv.waitKey(0)

class EdgeDetector:
    # Threshold for circularity (circle = 1); to be determined
    # Circularity = 4 * PI * AREA / (PERIMETER ^ 2)
    # In general, decrease this more bigger images
    # Based on test values of 0.72, 0.8, 0.68 seems like 0.4 is a good value
    # Threshold for concavity; the formula is area_of_blob / area_convexity_hull
    # Based on values around 0.95 and 0.98

    def __init__(self, circle_threshold = 0.4, convex_threshold = 0.85, min_inertia = 3, max_intertia = 5.5, percentage_led_area = 0.01):
        self.circle_threshold = circle_threshold
        self.convex_threshold = convex_threshold
        self.min_inertia = min_inertia
        self.max_inertia = max_intertia
        self.percentage_led_area = percentage_led_area

    # Finds the contours of the LEDs and returns them
    def get_edges(self, black_white_image, area = False, distance = False, already_grayscale = False):

        contours, hierarchy = cv.findContours(black_white_image, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE)
        min_area = black_white_image.shape[0] * black_white_image.shape[1] * self.percentage_led_area
        # Note : other versions of OpenCV will require image, contours, hierarchy
        # Now for the characteristics of the contours
        retained_contours = []
        for contour in contours:
            # TODO : something wrong here, probably with area
            area = cv.contourArea(contour)
            if area >= min_area:
                # Circularity < threshold : is the contour NOT circular?
                circularity = 4*PI*area / (cv.arcLength(contour, True) * cv.arcLength(contour, True))
                # Convexity > threashold : we want the shape to be convex
                convexity = area / cv.contourArea(cv.convexHull(contour))
                # Opencv docs aren't very helpful on this but looks like inertia is aspect ratio
                (x,y),(width,height),theta = cv.minAreaRect(contour)
                # This is basically aspect ratio
                inertia = width / height
                if inertia < 1:
                    inertia = 1/inertia
                # Note : the angle is pretty much useless
                logger.debug("calculated circularity : ", circularity)
                logger.debug("threshold  circularity : ", self.circle_threshold)
                logger.debug("\n")
                logger.debug("calculated converxity  : ", convexity)
                logger.debug("threshold  converxity  : ", self.convex_threshold)
                logger.debug("\n")
                logger.debug("calculated inertia     : ", inertia)
                logger.debug("threshold  inertia     : ", self.min_inertia, self.max_inertia)
                logger.debug("\n")
                logger.debug("image area           : ", black_white_image.shape[0] * black_white_image.shape[1])
                logger.debug("calculated area        : ", area)
                logger.debug("\n")
                logger.debug("detected angle         : ", theta)
                logger.debug("displaying contour, press any key to continue ... ")
                debug_contours([contour], black_white_image)
                # Make it able to switch on and off (ex. do not consider circularity)
                if circularity > self.circle_threshold and \
                    convexity >  self.convex_threshold and \
                    inertia > self.min_inertia and \
                    inertia < self.max_inertia:
                    retained_contours.append(contour)
                    logger.debug("retained !")
                else:
                    logger.debug("rejected !")
                logger.debug("----------------------------------")
                # Use HSV instead
                # 1. color = red or blue
                # Use the other 2 to see if its gray
        return retained_contours

if __name__ == '__main__':
    import os
    logger.setLevel("DEBUG") # Debug, info, warning, erreur
    image = cv.imread("test11.PNG")
    gbd = EdgeDetector()
    logger.debug(len(gbd.find_contours(image)))